#include <tcl.h>
#include "defs.h"
#include "Comm.h"
#include "Database.h"
#include "Synch.h"
#include "table.h"
#include "User.h"
#include "Random.h"
#include "Handle.h"
#include "assert.h"

unsigned long pc_id;  /* hotsync manager expects this as extern! */
extern char home_dir[256];
extern char username[256];
extern int daemon_flag;


/****** JetSync API function ******/
/* Initializes the communication between Linux and the PalmPilot */
/* Arguments: none */
/* Returns: via param, the inited pilot socket reference */
/**********************************/
int js_InitComm(int *sd)
{
  struct pi_sockaddr addr;
  int ret;

/* Pilot SOCKET */
  strcpy(addr.pi_device,getenv("PILOTPORT"));

  if (!(*sd = pi_socket(PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP))) {
    perror("pi_socket");
    return -1;
  }
    
  addr.pi_family = PI_AF_SLP;
  
  ret = pi_bind(*sd, (struct sockaddr*)&addr, sizeof(addr));
  if (ret == -1) {
    perror("pi_bind");
    return -1;
  }

  ret = pi_listen(*sd,1);
  if (ret == -1) {
    perror("pi_listen");
    return -1;
  }

  *sd = pi_accept(*sd, 0, 0);
  if (*sd == -1) {
    perror("pi_accept");
    return -1;
  }

  pierror(dlp_OpenConduit(*sd));
  return 0;
}


/* Action: Opens a connection with the pilot */
/*  Notes: Sets username if first pilot sync */
/*         Otherwise compares this user with the pilot's owner for */
/*         authenticity */
int js_Start(int *sd, struct PilotUser *U, char *name, char *path)
{
  int err;
  long userID;
  char path2[256];
  char pass[80];
  char tcl_string[256];
  DIR *d;


/* attempt to start communication with pilot */
  if (js_InitComm(sd) < 0) return -1;

/* USER info */
  if ((err = dlp_ReadUserInfo(*sd, U)) < 0) 
    printf ("Userinfo error: %d\n",err);


/* Does the pilot have a username assigned? */
  if (!strcmp(U->username,"")) {
    long id;

    printf("Establishing new username for pilot (%s)\n",name);

    err = js_get_user(name, &id, pass);
    if (err) {
/* in case of the daemon, since Tcl/Tk cannot be used, abort  */
      if (daemon_flag != 2) {
#ifdef __DEBUG
        printf("Cannot continue synchronization. Change to manual mode.\n");
#endif
        dlp_AddSyncLogEntry(*sd, "Cannot continue synchronization. Change to manual mode.");
        return -1;
      }

      U->userID = userID = abs(getrandom(ULONG_MAX));
      U->viewerID = 77;

      strcpy(g_interp->result,"");
      while (!strcmp(g_interp->result,"")) {
        sprintf(tcl_string, "show_info_dialog \"%s\"", name);
        Tcl_Eval(g_interp, tcl_string);
      }
      if (!strcmp(username,"NULL"))
        strcpy(username, strtok(g_interp->result,":"));

      js_ChangeUser(*sd, &U, name);
      add_user(U, strtok(NULL,":"));
      change_user(username);
    }
    else {
      U->userID = id;
      js_ChangeUser(*sd, &U, name);
    }
  }
  else { /* pilot has a name, check if it exists locally */
    long id;

    err = js_get_user(U->username, &id, pass);

    if (err) {
/* in case of the daemon, since Tcl/Tk cannot be used, abort  */
      if (daemon_flag != 2) {
#ifdef __DEBUG
        printf("Cannot continue synchronization. Change to manual mode.\n");
#endif
        dlp_AddSyncLogEntry(*sd, "Cannot continue synchronization. Change to manual mode.");
        return -1;
      }
    }

    strcpy(username, U->username);
    change_user(username);

/* if no such user or pilot user != local user, add user */
    if (err) {
      sprintf(tcl_string, "show_info_dialog \"%s\"", U->username);
      Tcl_Eval(g_interp, tcl_string);

      strtok(g_interp->result,":");
      add_user(U, strtok(NULL,":"));
      change_user(U->username);
    }
  }

  strcpy(path, home_dir);

  if (!strcmp(username,"NULL"))
    ASSERT(0); /* fatal error - should never be reached */

  /* check for existence of user directory */
  d = opendir(path);

  if (d)
    closedir(d);
  else {
    printf("JetSync was incorrectly installed. No '.jetsync' directory!\n");
    exit(-1);
  }

  sprintf(path2,"%s/%s", path, username);
  d = opendir(path2);
  if (!d) {
    FILE *fp;

    printf("Preparing a new jetsync directory for %s\n", U->username);
    if (mkdir(path2, 00700))
      printf("Could not create user directory for %s\n", U->username);

    strcat(path2,"/conf");
    if (mkdir(path2, 00700))
      printf("Could not create user conf directory for %s\n", U->username);

    fp=fopen(join_str(path2,"/prefs"),"w");
    fprintf(fp,"MODNUM:0\n");
    fclose(fp);

    fp=fopen(join_str(path2,"/jetmemo.dat"),"w");
    fprintf(fp,"%s/%s\n",path,U->username);
    fclose(fp);
    fp=fopen(join_str(path2,"/jettodo.dat"),"w");
    fprintf(fp,"%s/%s\n",path,U->username);
    fclose(fp);
    fp=fopen(join_str(path2,"/jetaddr.dat"),"w");
    fprintf(fp,"%s/%s\n",path,U->username);
    fclose(fp);
    fp=fopen(join_str(path2,"/jetdate.dat"),"w");
    fprintf(fp,"%s/%s\n",path,U->username);
    fclose(fp);
    fp=fopen(join_str(path2,"/jetmail.dat"),"w");
    fclose(fp);
    fp=fopen(join_str(path2,"/ical.map"),"w");
    fclose(fp);

    copy_default_file(join_str(path,"/default/color.def"),
                      join_str(path2,"/color.def"));
    copy_default_file(join_str(path,"/default/conduits.def"),
                      join_str(path2,"/conduits.def"));
    copy_default_file(join_str(path,"/default/conduits.def"),
                      join_str(path2,"/conduits.tmp"));

    copy_default_file(join_str(path,"/default/jetmemo.cat"),
                      join_str(path2,"/jetmemo.cat"));
    copy_default_file(join_str(path,"/default/jettodo.cat"),
                      join_str(path2,"/jettodo.cat"));
    copy_default_file(join_str(path,"/default/jetaddr.cat"),
                      join_str(path2,"/jetaddr.cat"));
    copy_default_file(join_str(path,"/default/jetdate.cat"),
                      join_str(path2,"/jetdate.cat"));


    sprintf(path2,"%s/%s", path, U->username);
    copy_default_file(join_str(path,"/default/memos"),
                      join_str(path2,"/memos"));
    copy_default_file(join_str(path,"/default/todos"),
                      join_str(path2,"/todos"));
    copy_default_file(join_str(path,"/default/addrs"),
                      join_str(path2,"/addrs"));
    copy_default_file(join_str(path,"/default/i.cal"),
                      join_str(path2,"/i.cal"));

  }
  closedir(d);

  return 0;
}


/****** JetSync API function ******/
/* Connects to the pilot, retrieves all DB names and creator ids. */
/* Arguments: inited pilot socket reference */
/* Returns: via params, a database list and its size */
/**********************************/
void js_Connect(int sd, struct DBList **ret, int *size)
{
  int i = 0, count = 0;
  struct DBInfo info;
  char tcl_string[256];

  *ret = NULL;

  sprintf(tcl_string, "progress 0 0 \"Retrieving database names...\"");
  Tcl_Eval(g_interp, tcl_string);


/* this loop could take a few seconds to finish because of the slow function
   dlp_ReadDBList */
  for (;;) {
    if (dlp_ReadDBList(sd, 0, 0x80, i, &info) < 0) break;

    i = info.index + 1;

    if (info.flags & dlpDBFlagResource) continue;

    *ret = (struct DBList *) realloc(*ret,
                                     (count + 1) * sizeof(struct DBList));

    strcpy((*ret)[count].name, info.name);
    (*ret)[count].creator = info.creator;
    (*ret)[count].flags = info.flags;
    count++;
  }

  if (daemon_flag == 2) {
/* make room for the two special conduits Install and Backup */
    *ret = (struct DBList *) realloc(*ret,
                                     (count+2) * sizeof(struct DBList));

    strcpy((*ret)[count].name, "Install");
    (*ret)[count].creator = str2long("inst");
    (*ret)[count].flags = 0;
    count++;
    strcpy((*ret)[count].name, "Backup");
    (*ret)[count].creator = str2long("bakk");
    (*ret)[count].flags = 0;
    count++;
  }


  sprintf(tcl_string, "progress 0 0 \"Synchronizing...\"");
  Tcl_Eval(g_interp, tcl_string);

  *size = count;
}


/* gets the preferences for a specified conduit (by creator) */
/* returns: synchType, procName */
int get_db_from_file(unsigned char *creator, char *synchType, char *proc)
{
  char path[256];
  FILE *fp;
  char line[80];

  if (!strcmp(username,"NULL")) return 0;

  sprintf(path,"%s/%s/%s",home_dir,username,JS_CONDUITS_TMP);
  fp = fopen(path,"r");

  if (!fp) printf("Couldn't open conduits file (%s)...\n", JS_CONDUITS_TMP);

  while (!feof(fp)) {
    fgets(line,80,fp);

    if (feof(fp)) break;

    strtok(line,":");
    strcpy(proc, strtok(NULL,":"));
    *synchType = '0' + atoi(strtok(NULL,":"));
    strtok(NULL,":");

    if (!strcmp(creator, strtok(NULL,":"))) {
      fclose(fp);
      return 0;
    }
  }
  fclose(fp);
  return -1;
}


/****** JetSync API function ******/
/* Performs global synch - checks if a conduit exists for every entry */
/* in the DB list */
/* Arguments: inited pilot socket reference, user data, database list, */
/*            its size, and the log buffer */
/* Returns: 0 or -1 in case of error */
/**********************************/
int js_Synch(sd, U, list, size, path, logbuffer)
int sd;
struct PilotUser *U;
struct DBList *list;
int size;
char *path;
char *logbuffer;
{
/* param size = total number of conduits to execute */
  int i, j, fast_synch = 0;
  int err, code, perc;
  char synchType = DO_NOTHING;
  char tcl_string[256];
  char proc[80];

/* read the conduit table from file - conduits are all hardwired in table.c */
  init_conduit_table();
  
  setup_conduit_table();

#ifdef __DEBUG
  printf("\033[1;33mStarting Synchronization... Available conduits in Pilot are:\n");
#endif

#ifdef __DEBUG
  for (i = 0; i < size; i++) {
    printf("%s  (%s)\n",list[i].name,
            long2str(list[i].creator));
  }
#endif



  perc=0;
  for (i = 0; i < size; i++) {
    perc = (2 * i * 100) / size;
    sprintf(tcl_string, "progress %d 0 \"Synchronizing %s...\"",
            perc,list[i].name);
    code = Tcl_Eval(g_interp, tcl_string);

    err = get_db_from_file(long2str(list[i].creator),&synchType, proc);
    if (err == -1) continue; /* no conduit for this DB entry */

    list[i].synchType = synchType;

  /* Determine if fast/slow synch
     1) Last synch was done with this PC
     2) Fast synch only considered in bidirectional synch */
    fast_synch = (U->lastSyncPC == pc_id) && (synchType == BIDIREC);

#ifdef __DEBUG
    printf ("\033[1;31m%s synch.\033[0m\n",fast_synch?"FAST":"SLOW");
#endif


    strcpy(list[i].appname, proc);


#ifdef __DEBUG
    printf("\033[1;32mConduit: %s\n\tproc = %s\tsynchType=%c\033[0m\n",
           list[i].name, list[i].appname, list[i].synchType);
#endif
 

/* synchType = 4  =>  Do Nothing. */
    if (synchType == DO_NOTHING) {  /* && bit_b*/
      char str[80];

      synchType = 'b';
      sprintf(str,"NOP %s", list[i].name);
      if (list[i].name[strlen(list[i].name)-2]=='D' &&
          list[i].name[strlen(list[i].name)-1]=='B')
        strcpy(str + strlen(str) - 2, "\n");
      else
        strcpy(str + strlen(str), "\n");
      strcat(logbuffer,str);
      continue;
    }


#ifdef __DEBUG
    printf("--------------------------------------------------------------
\033[1;36mExecuting conduit...\033[0m\n");
#endif

/*
* Call appropriate conduit **/

    if ((j = lookup_conduit_table(proc)) != -1) { /* lookup conduit by name */
      /* it exists -> execute it */
      __CALL__ conduit_table[j](sd, fast_synch, list[i].synchType,
                                path, logbuffer);
    }
    else { /* conduit is not installed */
      printf("Conduit is not installed!\n"); /* add to log */
    }


#ifdef __DEBUG
    printf("\033[1;36mExecuted conduit...\033[0m
--------------------------------------------------------------\n");
#endif

  } /* foreach conduit */

#ifdef __DEBUG
  printf("No more conduits to update\n");
#endif


  sprintf(tcl_string, "progress %d %d \"HotSync complete\"",100*2,100*2);
  code = Tcl_Eval(g_interp, tcl_string);

  return 0;
}


int logbook(char *logbuffer)
{
  FILE *fp;
  char path[256];
  time_t t;

  sprintf(path,"%s/%s/hotsync.log",home_dir,username);
  fp = fopen(path,"a");
  if (!fp) return -1;

  t = time(0);
  fprintf(fp,"Last HotSync: %s\n%s\n\n",ctime(&t),logbuffer);

  fclose(fp);
  return 0;
}


/****** JetSync API function ******/
/* Close the connection to the pilot */
/* Arguments: inited pilot socket reference, user data and the log buffer */
/* Returns: void */
/**********************************/
void js_CloseComm(int sd, struct PilotUser *U, char *logbuffer)
{
  int err;

/* After synch */
  U->lastSyncPC = pc_id;
  U->lastSyncDate = time(0);
  dlp_WriteUserInfo(sd, U);

  dlp_AddSyncLogEntry(sd, logbuffer);
  logbook(logbuffer);
  err = pi_close(sd);
  if (err) {
#ifdef __DEBUG
    printf("pi_close returned an error;  file: Comm.c line: %d\n",__LINE__);
#endif
  }
  return;
}
